# -*- coding: utf-8 -*-

"""
    This module is a REQUIRED extension to record the time of Features, Scenarios and Steps
"""

from datetime import datetime

from radish.hookregistry import after, before

__REQUIRED__ = True


@before.each_feature  # pylint: disable=no-member
def time_recorder_before_each_feature(feature):
    """
        Sets the starttime of the feature
    """
    feature.starttime = datetime.now()


@before.each_scenario  # pylint: disable=no-member
def time_recorder_before_each_scenario(scenario):
    """
        Sets the starttime of the scenario
    """
    scenario.starttime = datetime.now()


@before.each_step  # pylint: disable=no-member
def time_recorder_before_each_step(step):
    """
        Sets the starttime of the step
    """
    step.starttime = datetime.now()


@after.each_feature  # pylint: disable=no-member
def time_recorder_after_each_feature(feature):
    """
        Sets the endtime of the feature
    """
    feature.endtime = datetime.now()


@after.each_scenario  # pylint: disable=no-member
def time_recorder_after_each_scenario(scenario):
    """
        Sets the endtime of the scenario
    """
    scenario.endtime = datetime.now()


@after.each_step  # pylint: disable=no-member
def time_recorder_after_each_step(step):
    """
        Sets the endtime of the step
    """
    step.endtime = datetime.now()



# -*- coding: utf-8 -*-

"""
    This module provides a hook which generates a BDD XML result file at the end of the run.
"""

from getpass import getuser
from socket import gethostname
from lxml import etree
from datetime import timedelta
import re

from radish.terrain import world
from radish.hookregistry import after
from radish.exceptions import RadishError
from radish.scenariooutline import ScenarioOutline
from radish.scenarioloop import ScenarioLoop
from radish.step import Step
import radish.utils as utils


def _get_element_from_model(what, model):
    """
        Create a etree.Element from a given model
    """
    duration = str(model.duration.total_seconds()) if model.starttime and model.endtime else ""
    return etree.Element(
        what,
        sentence=model.sentence,
        id=str(model.id),
        result=model.state,
        starttime=utils.datetime_to_str(model.starttime),
        endtime=utils.datetime_to_str(model.endtime),
        duration=duration,
        testfile=model.path
    )


def _strip_ansi(text):
    """
        Strips ANSI modifiers from the given text
    """
    pattern = re.compile("(\\033\[\d+(?:;\d+)*m)")
    return pattern.sub("", text)


def generate_bdd_xml(features):
    """
        Generates the bdd xml
    """
    if not features:
        raise RadishError("No features given to generate BDD xml file")

    duration = timedelta()
    for feature in features:
        if feature.state in [Step.State.PASSED, Step.State.FAILED]:
            duration += feature.duration

    testrun_element = etree.Element(
        "testrun",
        starttime=utils.datetime_to_str(features[0].starttime),
        endtime=utils.datetime_to_str(features[-1].endtime),
        duration=str(duration.total_seconds()),
        agent="{}@{}".format(getuser(), gethostname())
    )

    for feature in features:
        if not feature.has_to_run(world.config.scenarios, world.config.feature_tags, world.config.scenario_tags):
            continue

        feature_element = _get_element_from_model("feature", feature)

        description_element = etree.Element("description")
        description_element.text = etree.CDATA("\n".join(feature.description))

        scenarios_element = etree.Element("scenarios")

        for scenario in (s for s in feature.all_scenarios if not isinstance(s, (ScenarioOutline, ScenarioLoop))):
            if not scenario.has_to_run(world.config.scenarios, world.config.feature_tags, world.config.scenario_tags):
                continue
            scenario_element = _get_element_from_model("scenario", scenario)

            for step in scenario.all_steps:
                step_element = _get_element_from_model("step", step)
                if step.state is Step.State.FAILED:
                    failure_element = etree.Element(
                        "failure",
                        type=step.failure.name,
                        message=step.failure.reason
                    )
                    failure_element.text = etree.CDATA(_strip_ansi(step.failure.traceback))
                    step_element.append(failure_element)
                scenario_element.append(step_element)
            scenarios_element.append(scenario_element)
        feature_element.append(description_element)
        feature_element.append(scenarios_element)
        testrun_element.append(feature_element)

    with open(world.config.bdd_xml, "w+") as f:
        f.write(etree.tostring(testrun_element, pretty_print=True, xml_declaration=True, encoding="utf-8"))


@after.all  # pylint: disable=no-member
def bdd_xml_writer_after_all(features, marker):  # pylint: disable=unused-argument
    """
        Generates a BDD XML file with the results
    """
    if not world.config.bdd_xml:
        return

    generate_bdd_xml(features)



# -*- coding: utf-8 -*-

"""
    This module provides some default ArgumentExpressions
"""

from radish.argexpregistry import arg_expr, ArgumentExpression


@arg_expr("Number", r"(\+|-)?\d+")
def arg_expr_number(text):
    """
        Argument Expression which expects a valid number

        :param str text: the text which was matched as number

        :returns: integer number
        :rtype: int
    """
    return int(text)


@arg_expr("FloatNumber", r"(\+|-)?\d+(\.\d+)?")
def arg_expr_floatnumber(text):
    """
        Argument Expression which expects a valid floating point number

        :param str text: the text which was matched as floating point number

        :returns: float number
        :rtype: float
    """
    return float(text)


@arg_expr("MathExpression", r"[0-9 +\-*/%.e]+")
def arg_expr_mathexpression(text):
    """
        Argument Expression which expects a valid math expression

        :param str text: the text which was matched as math expression

        :returns: calculated float number from the math expression
        :rtype: float
    """
    return float(eval(text))


@arg_expr("VariableName", r"[A-Za-z_][A-Za-z0-9_]*")
def arg_expr_variablename(text):
    """
        Argument Expression which expects a variable name

        :param str text: the text which was matched as variable name

        :returns: the variable name
        :rtype: str
    """
    return text



# -*- coding: utf-8 -*-

import os
import sys
from docopt import docopt
from time import time

from radish import __VERSION__
from radish.core import Core
from radish.loader import Loader
from radish.matcher import Matcher
from radish.stepregistry import StepRegistry
from radish.hookregistry import HookRegistry
from radish.runner import Runner
from radish.exceptions import FeatureFileNotFoundError, ScenarioNotFoundError, FeatureTagNotFoundError, ScenarioTagNotFoundError
from radish.errororacle import error_oracle, catch_unhandled_exception
from radish.terrain import world
import radish.utils as utils


def setup_config(arguments):
    """
        Parses the docopt arguments and creates a configuration object in terrain.world
    """
    world.config = lambda: None
    for key, value in arguments.items():
        config_key = key.replace("--", "").replace("-", "_").replace("<", "").replace(">", "")
        setattr(world.config, config_key, value)


def show_features(core):
    """
        Show the parsed features
    """
    # FIXME: load dynamically
    import radish.extensions.console_writer

    # set needed configuration
    world.config.write_steps_once = True
    if not sys.stdout.isatty():
        world.config.no_ansi = True

    runner = Runner(HookRegistry(), dry_run=True)
    runner.start(core.features_to_run, marker="show")
    return 0


def run_features(core):
    """
        Run the parsed features

        :param Core core: the radish core object
    """
    # FIXME: load dynamically
    import radish.extensions.argumentexpressions
    import radish.extensions.time_recorder
    import radish.extensions.syslog_writer
    import radish.extensions.console_writer
    import radish.extensions.endreport_writer
    import radish.extensions.failure_inspector
    import radish.extensions.failure_debugger
    import radish.extensions.bdd_xml_writer

    # set needed configuration
    world.config.expand = True

    # load user's custom python files
    loader = Loader(world.config.basedir)
    loader.load_all()

    # match feature file steps with user's step definitions
    Matcher.merge_steps(core.features, StepRegistry().steps)

    # run parsed features
    if world.config.marker == "time.time()":
        world.config.marker = int(time())

    # scenario choice
    amount_of_scenarios = sum(len(f.scenarios) for f in core.features_to_run)
    if world.config.scenarios:
        world.config.scenarios = [int(s) for s in world.config.scenarios.split(",")]
        for s in world.config.scenarios:
            if s <= 0 or s > amount_of_scenarios:
                raise ScenarioNotFoundError(s, amount_of_scenarios)

    # tags
    if world.config.feature_tags:
        world.config.feature_tags = [t for t in world.config.feature_tags.split(",")]
        for tag in world.config.feature_tags:
            if not any(f for f in core.features if tag in [t.name for t in f.tags]):
                raise FeatureTagNotFoundError(tag)

    if world.config.scenario_tags:
        world.config.scenario_tags = [t for t in world.config.scenario_tags.split(",")]
        for tag in world.config.scenario_tags:
            if not any(s for f in core.features for s in f.scenarios if tag in [t.name for t in s.tags]):
                raise ScenarioTagNotFoundError(tag)

    runner = Runner(HookRegistry(), early_exit=world.config.early_exit)
    return runner.start(core.features_to_run, marker=world.config.marker)


@error_oracle
def main():
    """
Usage:
    radish show <features>
           [--expand]
           [--no-ansi]
    radish <features>...
           [-b=<basedir> | --basedir=<basedir>]
           [-e | --early-exit]
           [--debug-steps]
           [--debug-after-failure]
           [--inspect-after-failure]
           [--bdd-xml=<bddxml>]
           [--no-ansi]
           [--no-line-jump]
           [--write-steps-once]
           [--write-ids]
           [-t | --with-traceback]
           [-m=<marker> | --marker=<marker>]
           [-p=<profile> | --profile=<profile>]
           [-d | --dry-run]
           [-s=<scenarios> | --scenarios=<scenarios>]
           [--shuffle]
           [--feature-tags=<feature_tags>]
           [--scenario-tags=<scenario_tags>]
    radish (-h | --help)
    radish (-v | --version)

Arguments:
    features                                    feature files to run

Options:
    -h --help                                   show this screen
    -v --version                                show version
    -e --early-exit                             stop the run after the first failed step
    --debug-steps                               debugs each step
    --debug-after-failure                       start python debugger after failure
    --inspect-after-failure                     start python shell after failure
    --bdd-xml=<bddxml>                          write BDD XML result file after run
    --no-ansi                                   print features without any ANSI sequences (like colors, line jump)
    --no-line-jump                              print features without line jumps (overwriting steps)
    --write-steps-once                          does not rewrite the steps (this option only makes sense in combination with the --no-ansi flag)
    --write-ids                                 write the feature, scenario and step id before the sentences
    -t --with-traceback                         show the Exception traceback when a step fails
    -m=<marker> --marker=<marker>               specify the marker for this run [default: time.time()]
    -p=<profile> --profile=<profile>            specify the profile which can be used in the step/hook implementation
    -b=<basedir> --basedir=<basedir>            set base dir from where the step.py and terrain.py will be loaded [default: $PWD/radish]
    -d --dry-run                                make dry run for the given feature files
    -s=<scenarios> --scenarios=<scenarios>      only run the specified scenarios (comma separated list)
    --shuffle                                   shuttle run order of features and scenarios
    --feature-tags=<feature_tags>               only run features with the given tags
    --scenario-tags=<scenario_tags>             only run scenarios with the given tags
    --expand                                    expand the feature file (all preconditions)

(C) Copyright 2013 by Timo Furrer <tuxtimo@gmail.com>
    """

    sys.excepthook = catch_unhandled_exception

    arguments = docopt("radish {}\n{}".format(__VERSION__, main.__doc__), version=__VERSION__)

    # store all arguments to configuration dict in terrain.world
    setup_config(arguments)

    core = Core()

    feature_files = []
    for given_feature in world.config.features:
        if not os.path.exists(given_feature):
            raise FeatureFileNotFoundError(given_feature)

        if os.path.isdir(given_feature):
            feature_files.extend(utils.recursive_glob(given_feature, "*.feature"))
            continue

        feature_files.append(given_feature)

    core.parse_features(feature_files)

    if not core.features:
        print("Error: no features given")
        return 1

    argument_dispatcher = [
        ((lambda: world.config.show), show_features),
        ((lambda: True), run_features)
    ]

    # radish command dispatching
    for to_run, method in argument_dispatcher:
        if to_run():
            return method(core)


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-

"""
    This radish extension module provide the functionality to write the end report
"""

# disable no-member lint error because of dynamic method from colorful
# pylint: disable=no-member

from datetime import timedelta
from colorful import colorful
from radish.hookregistry import after
from radish.step import Step
from radish.utils import console_write as write
from radish.scenariooutline import ScenarioOutline
from radish.scenarioloop import ScenarioLoop


@after.all  # pylint: disable=no-member
def console_write_after_all(features, marker):
    """
        Writes the endreport for all features

        :param list features: all features
    """
    stats = {
        "features": {"amount": 0, "passed": 0, "failed": 0, "skipped": 0, "untested": 0},
        "scenarios": {"amount": 0, "passed": 0, "failed": 0, "skipped": 0, "untested": 0},
        "steps": {"amount": 0, "passed": 0, "failed": 0, "skipped": 0, "untested": 0},
    }
    duration = timedelta()
    for feature in features:
        stats["features"]["amount"] += 1
        stats["features"][feature.state] += 1

        if feature.state in [Step.State.PASSED, Step.State.FAILED]:
            duration += feature.duration

        for scenario in feature.all_scenarios:
            if isinstance(scenario, ScenarioOutline):  # skip ScenarioOutlines
                continue
            if isinstance(scenario, ScenarioLoop):  # skip ScenarioLoop
                continue

            stats["scenarios"]["amount"] += 1
            stats["scenarios"][scenario.state] += 1
            for step in scenario.steps:
                stats["steps"]["amount"] += 1
                stats["steps"][step.state] += 1

    colored_closing_paren = colorful.bold_white(")")
    colored_comma = colorful.bold_white(", ")
    passed_word = colorful.bold_green("{} passed")
    failed_word = colorful.bold_red("{} failed")
    skipped_word = colorful.cyan("{} skipped")

    output = colorful.bold_white("{} features (".format(stats["features"]["amount"]))
    output += passed_word.format(stats["features"]["passed"])
    if stats["features"]["failed"]:
        output += colored_comma + failed_word.format(stats["features"]["failed"])
    if stats["features"]["skipped"]:
        output += colored_comma + skipped_word.format(stats["features"]["skipped"])
    output += colored_closing_paren

    output += "\n"
    output += colorful.bold_white("{} scenarios (".format(stats["scenarios"]["amount"]))
    output += passed_word.format(stats["scenarios"]["passed"])
    if stats["scenarios"]["failed"]:
        output += colored_comma + failed_word.format(stats["scenarios"]["failed"])
    if stats["scenarios"]["skipped"]:
        output += colored_comma + skipped_word.format(stats["scenarios"]["skipped"])
    output += colored_closing_paren

    output += "\n"
    output += colorful.bold_white("{} steps (".format(stats["steps"]["amount"]))
    output += passed_word.format(stats["steps"]["passed"])
    if stats["steps"]["failed"]:
        output += colored_comma + failed_word.format(stats["steps"]["failed"])
    if stats["steps"]["skipped"]:
        output += colored_comma + skipped_word.format(stats["steps"]["skipped"])
    output += colored_closing_paren

    output += "\n"
    output += colorful.cyan("Run {} finished within {}:{} minutes".format(marker, int(duration.total_seconds()) / 60, duration.total_seconds() % 60.0))

    write(output)



# -*- coding: utf-8 -*-

"""
    This module brings the benefit of variables to radish
"""

from radish.step import step
from radish.argexpregistry import ArgumentExpression


@step(ArgumentExpression(r"I set the variable {variable_name:VariableName} to {value:MathExpression}"))
def set_float_variable_step(step, variable_name, value):  # pylint: disable=redefined-outer-name
    """
        Sets the variable `variable_name` in the scenario context to the float value `value`

        :param str variable_name: the name of the variable
        :param float value: the value for the variable
    """
    setattr(step.context, variable_name, value)


@step(ArgumentExpression(r"I set the variable {variable_name:VariableName} to \"{value}\""))
def set_str_variable_step(step, variable_name, value):  # pylint: disable=redefined-outer-name
    """
        Sets the variable `variable_name` in the scenario context to the string `value`

        :param str variable_name: the name of the variable
        :param str value: the value for the variable
    """
    setattr(step.context, variable_name, value)



# -*- coding: utf-8 -*-

"""
    This module provides an extension to write all features, scenarios and steps to the syslog.
"""


from radish.terrain import world
from radish.feature import Feature
from radish.hookregistry import before, after

import syslog


def get_scenario_feature(scenario):
    """
        Gets the scenarios feature
    """
    if not isinstance(scenario.parent, Feature):
        return scenario.parent.parent

    return scenario.parent


def log(message):
    """
        Logs the given message to the syslog

        :param string message: the message to log
    """
    if isinstance(message, unicode):
        message = message.encode("utf8")
    syslog.syslog(syslog.LOG_INFO, message)


@before.all  # pylint: disable=no-member
def syslog_writer_before_all(features, marker):  # pylint: disable=unused-argument
    """
        Opens the syslog
    """
    syslog.openlog("radish")
    log(u"begin run {}".format(marker))


@after.all  # pylint: disable=no-member
def syslog_writer_after_all(features, marker):  # pylint: disable=unused-argument
    """
        Closes the syslog
    """
    log(u"end run {}".format(marker))
    syslog.closelog()


@before.each_feature  # pylint: disable=no-member
def syslog_writer_before_each_feature(feature):
    """
        Writes the feature to the syslog
    """
    log(u"begin feature {}:{} {}".format(world.config.marker, feature.id, feature.sentence))


@after.each_feature  # pylint: disable=no-member
def syslog_writer_after_each_feature(feature):
    """
        Writes the feature to the syslog
    """
    log(u"end feature {}:{} {}".format(world.config.marker, feature.id, feature.sentence))


@before.each_scenario  # pylint: disable=no-member
def syslog_writer_before_each_scenario(scenario):
    """
        Writes the scenario to the syslog
    """
    log(u"begin scenario {}:{}.{} {}".format(world.config.marker, get_scenario_feature(scenario).id, scenario.id, scenario.sentence))


@after.each_scenario  # pylint: disable=no-member
def syslog_writer_after_each_scenario(scenario):
    """
        Writes the scenario to the syslog
    """
    log(u"end scenario {}:{}.{} {}".format(world.config.marker, get_scenario_feature(scenario).id, scenario.id, scenario.sentence))


@before.each_step  # pylint: disable=no-member
def syslog_writer_before_each_step(step):
    """
        Writes the step to the syslog
    """
    log(u"begin step {}:{}.{}.{} {}".format(world.config.marker, get_scenario_feature(step.parent).id, step.parent.id, step.id, step.sentence))


@after.each_step  # pylint: disable=no-member
def syslog_writer_after_each_step(step):
    """
        Writes the step to the syslog
    """
    log(u"{} step {}:{}.{}.{} {}".format(step.state, world.config.marker, get_scenario_feature(step.parent).id, step.parent.id, step.id, step.sentence))



# -*- coding: utf-8 -*-


"""
    This module provides an extension which starts a debugger when a step fails
"""

from radish.terrain import world
from radish.hookregistry import after
from radish.step import Step
import radish.utils as utils


@after.each_step  # pylint: disable=no-member
def failure_debugger_after_each_step(step):
    """
        Starts a python debugger if the step failed
    """
    if not world.config.debug_after_failure or step.state is not Step.State.FAILED:
        return

    pdb = utils.get_debugger()
    pdb.set_trace()



# -*- coding: utf-8 -*-

"""
    This radish extension provides the functionality to write the feature file run to the console.
"""

# disable no-member lint error because of dynamic method from colorful
# pylint: disable=no-member

import os
from colorful import colorful

from radish.terrain import world
from radish.hookregistry import before, after
from radish.feature import Feature
from radish.scenariooutline import ScenarioOutline
from radish.scenarioloop import ScenarioLoop
from radish.step import Step
from radish.utils import console_write as write

__LAST_PRECONDITION__ = None


def get_color_func(state):
    """
        Returns the color func to use
    """
    if state == Step.State.PASSED:
        return colorful.bold_green
    elif state == Step.State.FAILED:
        return colorful.bold_red
    elif state:
        return colorful.cyan


def get_line_jump_seq():
    """
        Returns the line jump ANSI sequence
    """
    line_jump_seq = ""
    if not world.config.no_ansi and not world.config.no_line_jump and not world.config.write_steps_once:
        line_jump_seq = "\r\033[A\033[K"
    return line_jump_seq


def get_id_sentence_prefix(model, color_func, max_rows=None):
    """
        Returns the id from a model as sentence prefix

        :param Model model: a model with an id property
        :param function color_func: a function which gives coloring
        :param int max_rows: the maximum rows. Used for padding
    """
    padding = len("{}. ".format(max_rows)) if max_rows else 0
    return color_func("{1: >{0}}. ".format(padding, model.id)) if world.config.write_ids else ""


def get_id_padding(max_rows):
    """
        Returns the id padding
    """
    if not world.config.write_ids:
        return ""

    return " " * (len(str(max_rows)) + 2)


def get_table_col_widths(table):
    """
        Returns the width for every column of a table (lists in list)
    """
    return [max(len(str(col)) for col in row) for row in zip(*table)]  # pylint: disable=star-args


@before.each_feature  # pylint: disable=no-member
def console_writer_before_each_feature(feature):
    """
        Writes feature header to the console

        :param Feature feature: the feature to write to the console
    """
    output = ""
    for tag in feature.tags:
        output += colorful.cyan("@{}{}\n".format(tag.name, "({})".format(tag.arg) if tag.arg else ""))

    leading = "\n    " if feature.description else ""

    output += "{}{}: {}  # {}{}{}".format(
        get_id_sentence_prefix(feature, colorful.bold_cyan),
        colorful.bold_white(feature.keyword),
        colorful.bold_white(feature.sentence),
        colorful.bold_black(feature.path),
        leading,
        colorful.white("\n    ".join(feature.description))
    )
    write(output)


@before.each_scenario  # pylint: disable=no-member
def console_writer_before_each_scenario(scenario):
    """
        Writes the scenario header to the console

        :param Scenario scenario: the scenario to write to the console
    """
    output = "\n"
    if isinstance(scenario.parent, ScenarioOutline):
        if world.config.write_steps_once:
            return

        id_prefix = get_id_sentence_prefix(scenario, colorful.bold_brown, len(scenario.parent.scenarios))
        colored_pipe = colorful.bold_white("|")
        output = "        {0}{1} {2} {1}".format(
            id_prefix,
            colored_pipe,
            (" {} ").format(colored_pipe).join(
                colorful.bold_brown("{1: <{0}}".format(scenario.parent.get_column_width(i), x)) for i, x in enumerate(scenario.example.data)
            )
        )
    elif isinstance(scenario.parent, ScenarioLoop):
        if world.config.write_steps_once:
            return

        id_prefix = get_id_sentence_prefix(scenario, colorful.bold_brown, len(scenario.parent.scenarios))
        colored_pipe = colorful.bold_white("|")
        output = "        {0}{1} {2: <18} {1}".format(id_prefix, colored_pipe, colorful.bold_brown(scenario.iteration))
    else:
        id_prefix = get_id_sentence_prefix(scenario, colorful.bold_cyan)
        for tag in scenario.tags:
            if tag.name == "precondition" and world.config.expand and world.config.show:  # exceptional for show command when scenario steps expand and tag is a precondition -> comment it out
                output += colorful.white("    # @{}{}\n".format(tag.name, "({})".format(tag.arg) if tag.arg else ""))
            else:
                output += colorful.cyan("    @{}{}\n".format(tag.name, "({})".format(tag.arg) if tag.arg else ""))
        output += """    {}{}: {}""".format(id_prefix, colorful.bold_white(scenario.keyword), colorful.bold_white(scenario.sentence))
    write(output)


@before.each_step  # pylint: disable=no-member
def console_writer_before_each_step(step):
    """
        Writes the step to the console before it is run

        :param Step step: the step to write to the console
    """
    global __LAST_PRECONDITION__
    if not isinstance(step.parent.parent, Feature):
        return

    if world.config.write_steps_once:
        return

    output = ""
    if step.as_precondition and __LAST_PRECONDITION__ != step.as_precondition:
        output += colorful.white("      As precondition from {}: {}\n".format(os.path.basename(step.as_precondition.path), step.as_precondition.sentence))
    elif not step.as_precondition and __LAST_PRECONDITION__:
        output += colorful.white("      From scenario\n")

    __LAST_PRECONDITION__ = step.as_precondition
    output += "\r        {}{}".format(get_id_sentence_prefix(step, colorful.bold_brown), colorful.bold_brown(step.sentence))

    if step.text:
        id_padding = get_id_padding(len(step.parent.steps))
        output += colorful.bold_white('\n            {}"""'.format(id_padding))
        output += colorful.cyan("".join(["\n                {}{}".format(id_padding, l) for l in step.raw_text]))
        output += colorful.bold_white('\n            {}"""'.format(id_padding))

    if step.table:
        colored_pipe = colorful.bold_white("|")
        col_widths = get_table_col_widths(step.table)
        for row in step.table:
            output += "\n          {0} {1} {0}".format(colored_pipe, (" {} ").format(colored_pipe).join(
                colorful.bold_brown("{1: <{0}}".format(col_widths[i], x)) for i, x in enumerate(row)
            ))

    write(output)


@after.each_step  # pylint: disable=no-member
def console_writer_after_each_step(step):
    """
        Writes the step to the console after it was run

        :param Step step: the step to write to the console
    """
    if not isinstance(step.parent.parent, Feature):
        return

    color_func = get_color_func(step.state)
    line_jump_seq = get_line_jump_seq() * (((len(step.raw_text) + 3) if step.text else 1) + (len(step.table) if step.table else 0))
    output = "{}        {}{}".format(line_jump_seq, get_id_sentence_prefix(step, colorful.bold_cyan), color_func(step.sentence))

    if step.text:
        id_padding = get_id_padding(len(step.parent.steps))
        output += colorful.bold_white('\n            {}"""'.format(id_padding))
        output += colorful.cyan("".join(["\n                {}{}".format(id_padding, l) for l in step.raw_text]))
        output += colorful.bold_white('\n            {}"""'.format(id_padding))

    if step.table:
        colored_pipe = colorful.bold_white("|")
        col_widths = get_table_col_widths(step.table)
        for row in step.table:
            output += "\n          {0} {1} {0}".format(colored_pipe, (" {} ").format(colored_pipe).join(
                color_func("{1: <{0}}".format(col_widths[i], x)) for i, x in enumerate(row)
            ))

    if step.state == step.State.FAILED:
        if world.config.with_traceback:
            output += "\n          {}{}".format(get_id_padding(len(step.parent.steps) - 2), "\n          ".join([colorful.red(l) for l in step.failure.traceback.split("\n")[:-2]]))
        output += "\n          {}{}: {}".format(get_id_padding(len(step.parent.steps) - 2), colorful.bold_red(step.failure.name), colorful.red(step.failure.reason))

    write(output)


@after.each_scenario  # pylint: disable=no-member
def console_writer_after_each_scenario(scenario):
    """
        If the scenario is a ExampleScenario it will write the Examples header

        :param Scenario scenario: the scenario which was ran.
    """
    output = ""
    if isinstance(scenario, ScenarioOutline):
        output += "\n    {}:\n".format(colorful.bold_white(scenario.example_keyword))
        output += colorful.bold_white("        {}| {} |".format(
            get_id_padding(len(scenario.scenarios)),
            " | ".join("{1: <{0}}".format(scenario.get_column_width(i), x) for i, x in enumerate(scenario.examples_header))
        ))
    elif isinstance(scenario, ScenarioLoop):
        output += "\n    {}: {}".format(colorful.bold_white(scenario.iterations_keyword), colorful.cyan(scenario.iterations))
    elif isinstance(scenario.parent, ScenarioOutline):
        colored_pipe = colorful.bold_white("|")
        color_func = get_color_func(scenario.state)
        output += "{0}        {1}{2} {3} {2}".format(
            get_line_jump_seq(),
            get_id_sentence_prefix(scenario, colorful.bold_cyan, len(scenario.parent.scenarios)),
            colored_pipe,
            (" {} ").format(colored_pipe).join(
                color_func("{1: <{0}}".format(scenario.parent.get_column_width(i), x)) for i, x in enumerate(scenario.example.data)
            )
        )

        if scenario.state == Step.State.FAILED:
            failed_step = scenario.failed_step
            if world.config.with_traceback:
                output += "\n          {}{}".format(get_id_padding(len(scenario.parent.scenarios)), "\n          ".join([colorful.red(l) for l in failed_step.failure.traceback.split("\n")[:-2]]))
            output += "\n          {}{}: {}".format(get_id_padding(len(scenario.parent.scenarios)), colorful.bold_red(failed_step.failure.name), colorful.red(failed_step.failure.reason))
    elif isinstance(scenario.parent, ScenarioLoop):
        colored_pipe = colorful.bold_white("|")
        color_func = get_color_func(scenario.state)
        output += "{0}        {1}{2} {3: <18} {2}".format(get_line_jump_seq(), get_id_sentence_prefix(scenario, colorful.bold_cyan, len(scenario.parent.scenarios)), colored_pipe, color_func(scenario.iteration))

        if scenario.state == Step.State.FAILED:
            failed_step = scenario.failed_step
            if world.config.with_traceback:
                output += "\n          {}{}".format(get_id_padding(len(scenario.parent.scenarios)), "\n          ".join([colorful.red(l) for l in failed_step.failure.traceback.split("\n")[:-2]]))
            output += "\n          {}{}: {}".format(get_id_padding(len(scenario.parent.scenarios)), colorful.bold_red(failed_step.failure.name), colorful.red(failed_step.failure.reason))

    if output:
        write(output)


@after.each_feature  # pylint: disable=no-member
def console_writer_after_each_feature(feature):  # pylint: disable=unused-argument
    """
        Writes a newline after each feature

        :param Feature feature: the feature which was ran.
    """
    write("")



# -*- coding: utf-8 -*-

"""
    This module provides a registry for all custom steps which were decorated with the @step-decorator.
"""

import re
import inspect
import parse
from singleton import singleton

from radish.argexpregistry import ArgumentExpression
from radish.exceptions import RadishError, SameStepError, StepRegexError


@singleton()
class StepRegistry(object):
    """
        Represents the step registry
    """
    def __init__(self):
        self.clear()

    def register(self, regex, func):
        """
            Registers a given regex with the given step function.
        """
        if regex in self._steps:
            raise SameStepError(regex, self._steps[regex], func)

        self._steps[regex] = func

    def register_object(self, steps_object):
        """
            Registers all public methods from the given object as steps.
            The step regex must be in the first line of the docstring

            :param instance steps_object: the object with step definition methods
        """
        ignore = getattr(steps_object, "ignore", [])
        for attr in dir(steps_object):
            if attr in ignore or not inspect.ismethod(getattr(steps_object, attr)):  # attribute should be ignored
                continue

            step_definition = getattr(steps_object, attr)
            step_regex = self._extract_regex(step_definition)
            self.register(step_regex, step_definition)

        return steps_object

    def _extract_regex(self, func):
        """
            Extracts a step regex from the docstring of the given func

            :param function func: the step definition function
        """
        docstr = func.__doc__.strip()
        if not docstr:
            raise RadishError("Step definition '{}' from class must have step regex in docstring".format(func.__name__))

        regex = docstr.splitlines()[0]
        try:
            re.compile(regex)
        except re.error as e:
            raise StepRegexError(regex, func.__name__, e)

        return regex

    def clear(self):
        """
            Clears all registered steps
        """
        self._steps = {}

    @property
    def steps(self):
        """
            Returns all registered steps
        """
        return self._steps


def step(regex):
    """
        Step decorator for custom steps

        :param string regex: this is the regex to match the steps in the feature file

        :returns: the decorated function
        :rtype: function
    """
    def _decorator(func):
        """
            Represents the actual decorator
        """
        try:
            if not isinstance(regex, ArgumentExpression):
                re.compile(regex)
        except re.error as e:
            raise StepRegexError(regex, func.__name__, e)
        else:
            StepRegistry().register(regex, func)
        return func
    return _decorator


def steps(cls):
    """
        Decorator for classes with step definitions inside
    """
    old_cls_init = getattr(cls, "__init__")

    def new_cls_init(self, *args, **kwargs):
        """
            New __init__ method for the given class which calls the old __init__ method
            and registers the steps
        """
        old_cls_init(self, *args, **kwargs)

        # register functions as step
        StepRegistry().register_object(self)

    setattr(cls, "__init__", new_cls_init)
    return cls

given = lambda regex: step("Given {}".format(regex))  # pylint: disable=invalid-name
when = lambda regex: step("When {}".format(regex))  # pylint: disable=invalid-name
then = lambda regex: step("Then {}".format(regex))  # pylint: disable=invalid-name



# -*- coding: utf-8 -*-

"""
    This module provides an extension which starts a python shell after a step failed
"""

from radish.terrain import world
from radish.hookregistry import after
from radish.step import Step
from radish.exceptions import RadishError


@after.each_step  # pylint: disable=no-member
def failure_inspector_after_each_step(step):
    """
        Starts a python shell after a step failed
    """
    if not world.config.inspect_after_failure or step.state is not Step.State.FAILED:
        return

    try:
        from IPython import embed
    except ImportError as e:
        raise RadishError("Cannot import IPython embed function: {}".format(e))

    embed()



# -*- coding: utf-8 -*-

"""
    This module provides a class to match the feature file steps with the registered steps from the registry
"""

import re
import parse

from radish.argexpregistry import ArgExpRegistry, ArgumentExpression
from radish.exceptions import StepDefinitionNotFoundError, StepArgumentRegexError


class Matcher(object):
    """
        Matches steps from the feature files with the registered steps
    """
    @classmethod
    def merge_steps(cls, features, steps):
        """
            Merges steps from the given features with the given steps

            :param list features: the features
            :param dict steps: the steps
        """
        # FIXME: fix cycle-import ... Matcher -> ScenarioOutline -> Step -> Matcher
        from radish.scenariooutline import ScenarioOutline
        for feature in features:
            for scenario in feature.all_scenarios:
                if isinstance(scenario, ScenarioOutline):
                    continue  # ScenarioOutline steps do not have to be merged

                for step in scenario.steps:
                    cls.merge_step(step, steps)

    @classmethod
    def merge_step(cls, step, steps):
        """
            Merges a single step with the registered steps

            :param Step step: the step from a feature file to merge
            :param list steps: the registered steps
        """
        arguments, keyword_arguments, func = cls.match(step.expanded_sentence, steps)
        if not func:
            raise StepDefinitionNotFoundError(step)

        step.definition_func = func
        step.arguments = arguments
        step.keyword_arguments = keyword_arguments

    @classmethod
    def match(cls, sentence, steps):
        """
            Tries to find a match from the given sentence with the given steps

            :param string sentence: the step sentence to match
            :param dict steps: the available registered steps

            :returns: the arguments and the func which were matched
            :rtype: tuple
        """
        for regex, func in steps.items():
            if isinstance(regex, ArgumentExpression):
                try:
                    compiled = parse.compile(regex.regex, ArgExpRegistry().expressions)
                except ValueError as e:
                    raise StepArgumentRegexError(regex.regex, func.__name__, e)

                match = compiled.search(sentence)
                if match:
                    return match.fixed, match.named, func
            else:
                match = re.search(regex, sentence)
                if match:
                    return match.groups(), match.groupdict(), func

        return None, None, None



# -*- coding: utf-8 -*-

__VERSION__ = "0.2.5"

# export some functions for users
from radish.terrain import world
from radish.hookregistry import before, after
from radish.stepregistry import step, given, when, then, steps
from radish.argexpregistry import arg_expr, ArgumentExpression
from radish.exceptions import ValidationError



# -*- coding: utf-8 -*-

"""
    This module contains a class to load the step and terrain files
"""

import os
import sys
import fnmatch

import radish.utils as utils


class Loader(object):
    """
        Class to load a modules in a given folder
    """
    def __init__(self, location):
        self._location = os.path.expanduser(os.path.expandvars(location))
        self._loaded_modules = {}

    def load_all(self):
        """
            Loads all modules in the `location` folder
        """
        if not os.path.exists(self._location):
            raise OSError("Location '{}' to load modules does not exist".format(self._location))

        for p, d, f in os.walk(self._location):
            for filename in fnmatch.filter(f, "*.py"):
                self.load_module(os.path.join(p, filename))

    def load_module(self, path):
        """
            Loads a module by the given `path`

            :param string path: the path to the module to load
        """
        parent = os.path.dirname(utils.expandpath(path))
        sys.path.insert(0, parent)
        module_name = os.path.splitext(os.path.split(path)[1])[0]
        try:
            module = __import__(module_name)
        except Exception as e:
            #raise ImportError("Unable to import module '{}' from '{}': {}".format(module_name, path, e))
            raise e
        else:
            self._loaded_modules[module_name] = module
        finally:
            sys.path.remove(parent)
